(****** Homework 5 for CS-565 Programming Languages - Purdue. ******)
(** !!!Due date: 10/14/2016 by 9:00PM **)
(* All 1, 2, and 3 stars exercises have to be completed. 4 star *)
(* exercises are optional, but completing them can compensate for failed *)
(* exercises of this and future homeworks. *)

(* Submitted by: FILL IN YOUR NAME HERE *)

Require Import Coq.Bool.Bool.
Require Import Coq.Arith.Arith.
Require Import Coq.Arith.EqNat.
Require Import Coq.omega.Omega.
Require Import Maps.
Require Import Smallstep.
Require Import SfLib.
Require Import Stlc.
Require Import StlcProp.

(* Simply Typed Lambda Calculus *)

Module STLC_Exercises.
  Import STLC.
  Import STLCProp.

(** **** Exercise: 2 stars, typing_example_3  *)
(** Formally prove the following typing derivation holds: *)
(**
       empty |- \x:Bool->B. \y:Bool->Bool. \z:Bool.
                   y (x z)
             \in T.
*)

Example typing_example :
  exists T,
    empty |-
      (tabs x (TArrow TBool TBool)
         (tabs y (TArrow TBool TBool)
            (tabs z TBool
               (tapp (tvar y) (tapp (tvar x) (tvar z)))))) \in
      T.
Proof with auto.
(* FILL IN HERE *) Admitted.

(** **** Exercise: 3 stars, typing_nonexample_3  *)
(** Prove that no well-typed term can be applied to itself.

    ~ (exists S, exists T,
          empty |- \x:S. x x \in T).
*)

Example typing_nonexample :
  ~ (exists S, exists T,
        empty |-
          (tabs x S
             (tapp (tvar x) (tvar x))) \in
          T).
Proof.
  (* FILL IN HERE *) Admitted.
(** [] *)

(** **** Exercise: 2 stars, type_soundness  *)
(** Put progress and preservation together and show that a well-typed
    term can _never_ reach a stuck state.  *)

Theorem soundness : forall t t' T,
  empty |- t \in T ->
  t ==>* t' ->
  ~(stuck t').
Proof.
  intros t t' T Hhas_type Hmulti. unfold stuck.
  intros [Hnf Hnot_val]. unfold normal_form in Hnf.
  induction Hmulti.
  (* FILL IN HERE *) Admitted.
(** [] *)

End STLC_Exercises.

Module STLC_Extensions.

  (* For the last half of this homework, you are going to augment the *)
  (* formalization of the simply-typed lambda calculus with two extensions: *)
  (* let bindings, which we covered in class last week, *)
  (* and the unit base type, whose definitions follow. *)

  (* ================================================================= *)
  (** ** Let Bindings *)

  (** Syntax:

       t ::=                Terms
           | ...               (other terms same as before)
           | let x=t in t      let-binding
   *)

  (**
    Reduction:

                                 t1 ==> t1'
                     ----------------------------------               (ST_Let1)
                     let x=t1 in t2 ==> let x=t1' in t2

                        ----------------------------              (ST_LetValue)
                        let x=v1 in t2 ==> [x:=v1]t2

    Typing:

                Gamma |- t1 : T1      Gamma , x:T1 |- t2 : T2
                --------------------------------------------            (T_Let)
                        Gamma |- let x=t1 in t2 : T2
   *)

  (* ================================================================= *)
  (** ** Unit *)

  (** Another handy base type, found especially in languages in
    the ML family, is the singleton type [Unit]. *)
  (** It has a single element -- the term constant [unit] (with a small
    [u]) -- and a typing rule making [unit] an element of [Unit].  We
    also add [unit] to the set of possible values -- indeed, [unit] is
    the _only_ possible result of reducing an expression of type
    [Unit]. *)

  (** Syntax:

       t ::=                Terms
           | unit              unit value
           | ...

       v ::=                Values
           | unit              unit
           | ...

       T ::=                Types
           | Unit              Unit type
           | ...

    Typing:

                         --------------------                          (T_Unit)
                         Gamma |- unit : Unit
   *)

  (* Exercise: 1 star, Update the value judgement of the STLC *)
  (* to include these extensions: *)
  Inductive ty : Type :=
  | TArrow : ty -> ty -> ty
  | TUnit : ty.

  Inductive tm : Type :=
  (* pure STLC *)
  | tvar : id -> tm
  | tapp : tm -> tm -> tm
  | tabs : id -> ty -> tm -> tm
  (* units *)
  | tunit : tm
  (* let *)
  | tlet : id -> tm -> tm -> tm
  (* i.e., [let x = t1 in t2] *).

  Inductive value : tm -> Prop :=
  | v_abs : forall x T11 t12,
      value (tabs x T11 t12)
  (* FILL IN HERE.*).

  Hint Constructors value.

  (* Exercise: 2 star, Update the STLC substitution function and reduction rules as well *)

  (* ----------------------------------------------------------------- *)
  (** *** Substitution *)

  Fixpoint subst (x:id) (s:tm) (t:tm) : tm :=
    match t with
    | tvar y =>
      if beq_id x y then s else t
    | tabs y T t1 =>
      tabs y T (if beq_id x y then t1 else (subst x s t1))
    | tapp t1 t2 =>
      tapp (subst x s t1) (subst x s t2)
    (* FILL IN HERE *)
    | _ => t (* ... and delete this line. *)
    end.

  Notation "'[' x ':=' s ']' t" := (subst x s t) (at level 20).

  Inductive step : tm -> tm -> Prop :=
  | ST_AppAbs : forall x T11 t12 v2,
      value v2 ->
      (tapp (tabs x T11 t12) v2) ==> [x:=v2]t12
  | ST_App1 : forall t1 t1' t2,
      t1 ==> t1' ->
      (tapp t1 t2) ==> (tapp t1' t2)
  | ST_App2 : forall v1 t2 t2',
      value v1 ->
      t2 ==> t2' ->
      (tapp v1 t2) ==> (tapp v1 t2')
  (* let and unit *)
  (* FILL IN HERE *)

  where "t1 '==>' t2" := (step t1 t2).

  Notation multistep := (multi step).
  Notation "t1 '==>*' t2" := (multistep t1 t2) (at level 40).

  Hint Constructors step.

  (** Exercise, 2 stars: Update the typing rules according to the definitions *)
  (* from above as well. *)

  Definition context := partial_map ty.

  (** Next we define the typing rules.  These are nearly direct
    transcriptions of the inference rules shown above. *)

  Reserved Notation "Gamma '|-' t '\in' T" (at level 40).

  Inductive has_type : context -> tm -> ty -> Prop :=
  (* Typing rules for proper terms *)
  | T_Var : forall Gamma x T,
      Gamma x = Some T ->
      Gamma |- (tvar x) \in T
  | T_Abs : forall Gamma x T11 T12 t12,
      (update Gamma x T11) |- t12 \in T12 ->
                                      Gamma |- (tabs x T11 t12) \in (TArrow T11 T12)
  | T_App : forall T1 T2 Gamma t1 t2,
      Gamma |- t1 \in (TArrow T1 T2) ->
                      Gamma |- t2 \in T1 ->
                                      Gamma |- (tapp t1 t2) \in T2
  (* unit + let *)
  (* FILL IN HERE *)
  where "Gamma '|-' t '\in' T" := (has_type Gamma t T).

  Hint Constructors has_type.

  Module Examples.
    (* Exercise, 1 star: Validate your definitions using an example. *)
    Hint Extern 2 (has_type _ (tapp _ _) _) =>
    eapply T_App; auto.

    (* Notations for variable identifiers. *)
    Notation f := (Id 0).
    Notation x := (Id 1).

    (* let f = \x. x in f (f unit) *)
    Definition test :=
      tlet
        f
        (tabs x TUnit (tvar x))
        (tapp (tvar f) (tapp (tvar f) tunit)).
    (* Uncomment these lines. *)
    Example typechecks :
      (@empty ty) |- test \in TUnit.
    Proof. Admitted. (* unfold test. eauto 15. Qed. *)

    Example reduces :
      test ==>* tunit.
    Proof. Admitted. (* unfold test. normalize. Qed. *)

  End Examples.

  (** Exercise, 3 stars: Update the proof of progress for the extended language *)

  Theorem progress : forall t T,
     empty |- t \in T ->
     value t \/ exists t', t ==> t'.
  Proof with eauto.
    (* Theorem: Suppose empty |- t : T.  Then either
       1. t is a value, or
       2. t ==> t' for some t'.
     Proof: By induction on the given typing derivation. *)
    intros t T Ht.
    remember (@empty ty) as Gamma.
    generalize dependent HeqGamma.
    induction Ht; intros HeqGamma; subst.
    - (* T_Var *)
      (* The final rule in the given typing derivation cannot be
       [T_Var], since it can never be the case that
       [empty |- x : T] (since the context is empty). *)
      inversion H.
    - (* T_Abs *)
      (* If the [T_Abs] rule was the last used, then
       [t = tabs x T11 t12], which is a value. *)
      left...
    - (* T_App *)
      (* If the last rule applied was T_App, then [t = t1 t2],
       and we know from the form of the rule that
         [empty |- t1 : T1 -> T2]
         [empty |- t2 : T1]
       By the induction hypothesis, each of t1 and t2 either is
       a value or can take a step. *)
      right.
      destruct IHHt1; subst...
      + (* t1 is a value *)
        destruct IHHt2; subst...
        * (* t2 is a value *)
          (* If both [t1] and [t2] are values, then we know that
           [t1 = tabs x T11 t12], since abstractions are the
           only values that can have an arrow type.  But
           [(tabs x T11 t12) t2 ==> [x:=t2]t12] by [ST_AppAbs]. *)
          inversion H; subst; try (solve by inversion).
          exists (subst x t2 t12)...
        * (* t2 steps *)
          (* If [t1] is a value and [t2 ==> t2'],
           then [t1 t2 ==> t1 t2'] by [ST_App2]. *)
          inversion H0 as [t2' Hstp]. exists (tapp t1 t2')...
      + (* t1 steps *)
        (* Finally, If [t1 ==> t1'], then [t1 t2 ==> t1' t2]
         by [ST_App1]. *)
        inversion H as [t1' Hstp]. exists (tapp t1' t2)...
    (* - T_Unit *)
    (* FILL IN HERE *)
    (* - T_Let let *)
    (* FILL IN HERE *)
  Qed.

  (* Exercise, 2 stars: extend the definition of appears_free_in and *)
  (* the proofs of context invariance and free_in_context for the *)
  (* updated language *)

  (* This predicate establishes if a variable appears free in a term *)
  (* i.e. it's not "bound" by a lambda or let expression *)
  Inductive appears_free_in : id -> tm -> Prop :=
  | afi_var : forall x,
      appears_free_in x (tvar x)
  | afi_app1 : forall x t1 t2,
      appears_free_in x t1 -> appears_free_in x (tapp t1 t2)
  | afi_app2 : forall x t1 t2,
      appears_free_in x t2 -> appears_free_in x (tapp t1 t2)
  | afi_abs : forall x y T11 t12,
      y <> x  ->
      appears_free_in x t12 ->
      appears_free_in x (tabs y T11 t12)
  (* let *)
  (* FILL IN HERE *).

  Hint Constructors appears_free_in.

  (* This lemma states that a term remains well-typed, even when *)
  (* new (unused) variables are added to a typing context. *)
  Lemma context_invariance : forall Gamma Gamma' t S,
      Gamma |- t \in S  -> (* If t is well-typed in Gamma *)
      (forall x, appears_free_in x t -> Gamma x = Gamma' x)  ->
      (* and a new context Gamma' has the same values for all the free variables in t, *)
      Gamma' |- t \in S. (* t is well-typed in Gamma' *)
  Proof with eauto.
    intros. generalize dependent Gamma'.
    induction H;
      intros Gamma' Heqv...
    - (* T_Var *)
      apply T_Var... rewrite <- Heqv...
    - (* T_Abs *)
      apply T_Abs... apply IHhas_type. intros y Hafi.
      unfold update, t_update.
      destruct (beq_idP x y)...
    (* let *)
    (* FILL IN HERE *)
  Qed.

  (* This lemma states that every free variable x in a well-typed term t *)
  (* appears in the typing context. *)
  Lemma free_in_context : forall x t T Gamma,
      appears_free_in x t ->
      Gamma |- t \in T ->
      exists T', Gamma x = Some T'.
  Proof with eauto.
    intros x t T Gamma Hafi Htyp.
    induction Htyp; inversion Hafi; subst...
    - (* T_Abs *)
      destruct IHHtyp as [T' Hctx]... exists T'.
      unfold update, t_update in Hctx.
      rewrite false_beq_id in Hctx...
  Qed.

  (** Exercise, 2 stars: Update the proof  that substitution preserves typing *)
  (* for the extended language *)
  Lemma substitution_preserves_typing : forall Gamma x U v t S,
      (update Gamma x U) |- t \in S  ->
      empty |- v \in U   ->
      Gamma |- ([x:=v]t) \in S.
  Proof with eauto.
    (* Theorem: If Gamma,x:U |- t : S and empty |- v : U, then
     Gamma |- [x:=v]t : S. *)
    intros Gamma x U v t S Htypt Htypv.
    generalize dependent Gamma. generalize dependent S.
    (* Proof: By induction on the term t.  Most cases follow
     directly from the IH, with the exception of tvar
     and tabs. These aren't automatic because we must
     reason about how the variables interact. *)
    induction t;
      intros S Gamma Htypt; simpl; inversion Htypt; subst...
    - (* tvar *)
      simpl. rename i into y.
      (* If t = y, we know that
         [empty |- v : U] and
         [Gamma,x:U |- y : S]
       and, by inversion, [update Gamma x U y = Some S].  We want to
       show that [Gamma |- [x:=v]y : S].

       There are two cases to consider: either [x=y] or [x<>y]. *)
      unfold update, t_update in H1.
      destruct (beq_idP x y).
      + (* x=y *)
        (* If [x = y], then we know that [U = S], and that
         [[x:=v]y = v].  So what we really must show is
         that if [empty |- v : U] then [Gamma |- v : U].
         We have already proven a more general version
         of this theorem, called context invariance. *)
        subst.
        inversion H1; subst. clear H1.
        eapply context_invariance...
        intros x Hcontra.
        destruct (free_in_context _ _ S empty Hcontra)
          as [T' HT']...
        inversion HT'.
      + (* x<>y *)
        (* If [x <> y], then [Gamma y = Some S] and the substitution has no
       effect.  We can show that [Gamma |- y : S] by [T_Var]. *)
        apply T_Var...
    - (* tabs *)
      rename i into y. rename t into T11.
      (* If [t = tabs y T11 t0], then we know that
         [Gamma,x:U |- tabs y T11 t0 : T11->T12]
         [Gamma,x:U,y:T11 |- t0 : T12]
         [empty |- v : U]
       As our IH, we know that forall S Gamma,
         [Gamma,x:U |- t0 : S -> Gamma |- [x:=v]t0 : S].

       We can calculate that
         [x:=v]t = tabs y T11 (if beq_id x y then t0 else [x:=v]t0)
       And we must show that [Gamma |- [x:=v]t : T11->T12].  We know
       we will do so using [T_Abs], so it remains to be shown that:
         [Gamma,y:T11 |- if beq_id x y then t0 else [x:=v]t0 : T12]
       We consider two cases: [x = y] and [x <> y].
       *)
      apply T_Abs...
      destruct (beq_idP x y) as [Hxy|Hxy].
      + (* x=y *)
        (* If [x = y], then the substitution has no effect.  Context
       invariance shows that [Gamma,y:U,y:T11] and [Gamma,y:T11] are
       equivalent.  Since the former context shows that
       [t0 : T12], so does the latter. *)
        eapply context_invariance...
        subst.
        intros x Hafi. unfold update, t_update.
        destruct (beq_id y x)...
      + (* x<>y *)
        (* If [x <> y], then the IH and context invariance allow
         us to show that
           [Gamma,x:U,y:T11 |- t0 : T12]       =>
           [Gamma,y:T11,x:U |- t0 : T12]       =>
           [Gamma,y:T11 |- [x:=v]t0 : T12] *)
        apply IHt. eapply context_invariance...
        intros z Hafi. unfold update, t_update.
        destruct (beq_idP y z) as [Hyz|Hyz]...
        subst.
        rewrite false_beq_id...
    (* let *)
    (* FILL IN HERE *)
  Qed.


  (** Exercise, 3 stars: Update the proof of preservation for the extended language *)

  Theorem preservation : forall t t' T,
      empty |- t \in T  ->
      t ==> t'  ->
      empty |- t' \in T.
  Proof with eauto.
    intros t t' T HT.
    (* Theorem: If [empty |- t : T] and [t ==> t'], then
     [empty |- t' : T]. *)
    remember (@empty ty) as Gamma. generalize dependent HeqGamma.
    generalize dependent t'.
    (* Proof: By induction on the given typing derivation.  Many
     cases are contradictory ([T_Var], [T_Abs]).  We show just
     the interesting ones. *)
    induction HT;
      intros t' HeqGamma HE; subst; inversion HE; subst...
    - (* T_App *)
      (* If the last rule used was [T_App], then [t = t1 t2], and
       three rules could have been used to show [t ==> t']:
       [ST_App1], [ST_App2], and [ST_AppAbs]. In the first two
       cases, the result follows directly from the IH. *)
      inversion HE; subst...
      + (* ST_AppAbs *)
        (* For the third case, suppose
           [t1 = tabs x T11 t12]
         and
           [t2 = v2].
         We must show that [empty |- [x:=v2]t12 : T2].
         We know by assumption that
             [empty |- tabs x T11 t12 : T1->T2]
         and by inversion
             [x:T1 |- t12 : T2]
         We have already proven that substitution preserves
         typing, and
             [empty |- v2 : T1]
         by assumption, so we are done. *)
        apply substitution_preserves_typing with T1...
        inversion HT1...
    (* let *)
    (* FILL IN HERE *)
  Qed.

  Definition stuck (t:tm) : Prop :=
    (normal_form step) t /\ ~ value t.

  (* Exercise, 1 stars: Finally, prove soundness for the extended type system *)
  (* Hint: You should be able to copy and paste your proof from the previous section :) *)
  Theorem soundness : forall t t' T,
      empty |- t \in T ->
      t ==>* t' ->
      ~(stuck t').
  Proof.
  intros t t' T Hhas_type Hmulti. unfold stuck.
  intros [Hnf Hnot_val]. unfold normal_form in Hnf.
  induction Hmulti.
  (* FILL IN HERE *) Admitted.

End STLC_Extensions.
